/* The Zen kernel, GPL, azz@gnu.org

TO DO:

Filesystem length should be longs.

Filesystem flags and isatty, then fix up cat so that it can be used
for copying files (i.e. get it to check if its source is a tty and if
so check for the EOF character else check feof).

Pipes. Perhaps an external "pipe" command? Uses two temporary files.

Should fix up system()'s command-line splitting to cope
with quotes.

Copy/move commands in ed.

Sensible tracking for memory allocation. Implement PIDs, basically, so
that when a program exit()s (which it can't do, yet?) all its memory
gets freed, and free twice (or free rubbish) gets detected. (exit()
would be a setjmp/longjmp operation.)

Rewrite memcpy and memset in assembler for speed.

Perhaps fopen and diropen should increment the refcount of files and
do the equivalent of the unlink check on close? This is probably a bad
idea performance-wise (no, it's not!) on a single-task system, but is
required for multitasking. (Under UNIX, you can open a file, unlink it
and carry on using it, and it gets deleted when you close it. Few
programs do this, though.)

KNOWN BUGS:

fetchinode will fail when an inode overlaps a block boundary This
should be a fix in the InodeBlock* macros---rig it so that inodes
don't overlap boundaries. It's fine for dirents to overlap, though;
the rather hairy code handles it fine.
-> Note that inodes currently don't overlap block boundaries, as they
   are now 8 bytes, which divides nicely into the 1K block size.

File lengths are currently limited to 32K. Problem? Just move
INODE->len and FILE->len to longs. This is likely to be an issue if I
do a port to a non-8-bit architecture. (it breaks the sed executable at
the moment...)

Everything assumes infinite disk space and memory. Actually, I've
fixed (most) mallocs to check that they get memory, but other
functions are assumed to return a sensible value; ditto with
getblock() and freeblock(), which will print an error and return 0,
probably with disasterous consequences.

*/

#include "zen.h"
/* Number of [initial] disk buffers. 3 is plenty. */
#define NUMBUFFERS 4
/* Set to !=0 for paging, which is useful for debugging and on
   excessively small displays. */
#define PAGESIZE 0
/* Set to 1 for alloc tracking, which is useful for debugging. */
#define ALLOCTRACK 0
/* Set to 1 for disk tracking, which shows physical disk accesses. */
#define DISKTRACK 0
/* Set to 1 for buffer-usage tracking. */
#define BUFTRACK 0
/* Set to 1 for automatically bootstrapping the system via the serial port. */
#define BOOTSTRAP 1
/* Set to 1 for executable relocation tracking. */
#define RELOCTRACK 0


/* --- protos --- */

/* Are there too many damn prototypes here or what? Why couldn't they
   make C compilers clever enough to scan for prototypes from the
   function definitions first? It's the only thing that annoys me
   about C. */

/* Stuff in the C library. */
extern int strcmp(char *p, char *q);
extern char *index(char *s, char c);
extern char *rindex(char *s, char c);
extern int strncmp(char *p, char *q, int l);
extern int strlen(char *s);
extern int atoi(char *s);
extern char in(int p);
extern void out(int p,char c);

/* The following are in approximate chronological order. */
void main();
FILE *fopen(char *name, char *mode);
void fclose(FILE *file);
void init_os();
void fprintf(FILE *f, char *s, ...);
void printf(char *s, ...);
void doprintf(FILE *f, char *s, int *a);
char fputc(char c, FILE *f);
char fgetc(FILE *f);
int fputs(char *s, FILE *f);
int fputn(long n, FILE *f);
int fputu(unsigned int n, FILE *f);
char *malloc(unsigned nbytes);
void free(char *ap);
void readblock(int b, char *buf);
void writeblock(int b, char *buf);
void getblock(int b);
void sync();
int allocblock();
int mkfs();
int create(char *n, char t);
INODE *fetchinode(int n);
int allocinode(char t);
DIR *opendir(int i);
void closedir(DIR *d);
DIRENT *readdir(DIR *d);
int ls(char *n);
int nametoinode(char *name);
char *dirname(char *name);
char *basename(char *name);
int mkdir(char *name);
int mkdev(char *name, int dev, int unit);
int sh();
char *fgets(char *s, int n, FILE *f);
int mkdirentry(char *n, int i);
int unlink(char *name);
void memcpy(char *a, char *b, int c);
void freeblock(int b);
void freeinode(int i);
int link(char *old, char *new);
int feof(FILE *f);
int system(char *cmd);
int cat_main(int argc, char **argv);
void filecopy(FILE *f);
int dl_main(int argc,char **argv);
void fflush(FILE *f);
int chdir(char *s);

/* globals */

FILE *console;
extern FILE *stdin, *stdout, *stderr;
extern int ramstart;
int currentdir, parentinode, pagercount, pagercols;
BUFFER *currentbuffer;

/* code */

void main() {

  init_os();

#if BOOTSTRAP
  mkfs();
  system("dl >/bin/do");
  system("dl >t");
  system("do t");
#endif

  sh();

  fprintf(console,"System halt\n");
  halt();
}

#define MAXCMD 80

int sh() {
  char *cmdline;
  int rc;

  cmdline=malloc(MAXCMD);
  if (!cmdline) return FAILED;
  strcpy(cmdline,"help");
  do {
    rc=system(cmdline);
    if (rc!=NULL) printf("program returned exit status %d\n", rc);
    printf("(%u) Zen> ", memfree());
    sync(); /* Not strictly needed, but safer; like "sync" */
    fgets(cmdline, MAXCMD, stdin);
  } while (strcmp(cmdline,"exit\n"));
  free(cmdline);
  return NULL;
}

unsigned int freadint(f)
     FILE *f;
{
  unsigned char i;

  i=(unsigned char)fgetc(f);
  return (((unsigned char)fgetc(f))<<8)+i;
}

int loadbin(f,argc,argv)
     FILE *f;
     int argc;
     char **argv;
{
  unsigned int len, i, l, rc;
  unsigned char *prog, o;
  int (*newmain)();
  FILE *oi, *oo, *oe;

  if (freadint(f)==25981) { /* magic number check */
    len=freadint(f);
    prog=(unsigned char *)malloc(len);
#if RELOCTRACK
    fprintf(stderr,"loading %u bytes at %u\n",len,prog);
#endif
    if (prog) {
      for(i=0;i<len;i++) prog[i]=fgetc(f);
      i=freadint(f); /* number of patches */
      while(i--) { 
	l=freadint(f);
	o=prog[l];
	prog[l] += ((unsigned int)prog)&255;
	prog[l+1] += ((unsigned int)prog)>>8;
	if (prog[l]<o) {
	  prog[l+1]++;
	}
      }
      newmain=(void *)prog; /* only way to do it? */
#if RELOCTRACK
      fprintf(stderr,"before call: %u\n", memfree());
#endif
      oi=stdin; oo=stdout; oe=stderr; /* allow programs to modify */
      rc=(*newmain)(argc,argv);
      stdin=oi; stdout=oo; stderr=oe;
#if RELOCTRACK
      fprintf(stder,,"after call: %u\n", memfree());
#endif
      free((char *)prog);
      return rc;
    } else {
      fprintf(stderr,"insufficient memory to load\n");
      return FAILED;
    }
  } else {
    fprintf(stderr,"bad magic number\n");
    return FAILED;
  }
}

/* This is *very* different from how system() works on UNIX. Under
   UNIX, system() calls the default shell; here it *is* the shell.
*/
int system(cmd)
     char *cmd;
{
  FILE *origstdout, *origstdin, *origstderr, *f;
  int argc, c, rc, flag;
  char **argv, *s, *r;

  origstdout = origstdin = origstderr = NULL;
  if (cmd[strlen(cmd)-1]=='\n') cmd[strlen(cmd)-1]='\0';
  argc=1;
  for(s=cmd;*s;s++) if (*s==' ') argc++;
  argv=(char **)malloc(argc*(sizeof s));
  if (!argv) return FAILED;
  r=s=cmd;
  c=0;
  flag=1;
  while((s=index(s,' '))||flag) {
    if (s) *s++='\0';
    switch(r[0]) {
    case '>':
      origstdout = stdout;
      if (r[1]=='>') {
	stdout = fopen(r+2,"a");
      } else {
	stdout = fopen(r+1,"w");
      }
      if (!stdout) {
	fprintf(stderr,"unable to open output redirector %s\n",r);
	stdout=origstdout;
	origstdout=NULL;
	goto backout;
      }
      argc--;
      break;
    case '<':
      origstdin = stdin;
      stdin = fopen(r+1, "r");
      if (!stdin) {
	fprintf(stderr,"unable to open input redirector %s\n",r);
	stdin=origstdin;
	origstdin=NULL;
	if (origstdout) {
	  fclose(stdout);
	  stdout = origstdout;
	}
	goto backout;
      }
      argc--;
      break;
    default:
      argv[c++]=r;
      break;
    }
    if (s) {
      r=s;
    } else {
      flag=0;
    }
  }
  /* We now have argc and argv set up right. */
  rc=FAILED;
  r=malloc(6+strlen(argv[0]));
  if (!r) goto backout;
  strcpy(r,"/bin/");
  strcat(r,argv[0]);
  /* binaries replace builtins if possible */
  if (f=fopen(r,"r")) {
    rc=loadbin(f,argc,argv);
    fclose(f);
  } else {
    if (!strcmp(argv[0],"help")) {
      printf("builtins: ls mkdir mkdev touch ln rm mv cat dl cd mkfs sh\n");
      rc=NULL;
    }
    if (!strcmp(argv[0],"ls")) {
      if (argc==2) {
	rc=ls(argv[1]);
      } else {
	rc=ls(NULL);
      }
    }
    if ((!strcmp(argv[0],"mkdir"))&&(argc==2)) rc=mkdir(argv[1]);
    if ((!strcmp(argv[0],"mkdev"))&&(argc==4)) rc=mkdev(argv[1],atoi(argv[2]),atoi(argv[3]));
    if ((!strcmp(argv[0],"touch"))&&(argc==2)) rc=create(argv[1],IN_TYPE_FILE);
    if ((!strcmp(argv[0],"ln"))&&(argc==3)) rc=link(argv[1],argv[2]);
    if ((!strcmp(argv[0],"rm"))&&(argc==2)) rc=unlink(argv[1]);
    if ((!strcmp(argv[0],"mv"))&&(argc==3)) {
      /* safety: don't unlink unless the link was succesful */
      if (link(argv[1],argv[2])) {
	rc=FAILED;
      } else {
	rc=unlink(argv[1]);
      }
    }
    if (!strcmp(argv[0],"cat")) rc=cat_main(argc,argv);
    if (!strcmp(argv[0],"dl")) rc=dl_main(argc,argv);
    if ((!strcmp(argv[0],"cd"))&&(argc=2)) rc=chdir(argv[1]);
    if (!strcmp(argv[0],"mkfs")) rc=mkfs();
    if (!strcmp(argv[0],"sh")) rc=sh();
    if (!strcmp(argv[0],"sync")) {
      rc=NULL;
      sync();
    }
  }
  if (origstdout) {
    fclose(stdout);
    stdout = origstdout;
  }
  if (origstdin) {
    fclose(stdin);
    stdin = origstdin;
  }
  free(r);
 backout:
  free((char *)argv);
  return rc;
}

int chdir(path) 
char *path;
{
  int c;

  c=nametoinode(path);
  if (c) {
    currentdir=c;
    return NULL;
  } else {
    return FAILED;
  }
}

void init_os() {
  int i;
  BUFFER *b, *bb;

  console = fopen(NULL, "rw");
  stdin = stdout = stderr = console;
  printf("zen %s booting\n", VERSION);
  printf("console FILE* is at %u\n", console);

  b = bb = currentbuffer = NULL;
  for(i=0; i<NUMBUFFERS; i++) {
    if (!(b = (BUFFER *)malloc(sizeof(BUFFER)))) return;
    b->age = 0;
    b->block = i;
    b->next = bb;
    if (!(b->buf = malloc(BLOCKSIZE))) {
      printf("out of memory allocating disk buffers\n");
      return; /* and halt, presumably */
    }
    readblock(i, b->buf);
    if (!currentbuffer) currentbuffer = b;
    printf("allocated disk buffer @%u\n", (int)(b->buf));
    bb = b;
  }
  currentbuffer->next = b;
  currentdir = 1;
  pagercount = PAGESIZE;
  pagercols = 0;
}

int ls(n) 
     char *n;
{
  DIR *d;
  DIRENT *e;
  INODE *o;
  int i;

  if (n) {
    if (!(i=nametoinode(n))) {
      fprintf(stderr,"ls: no such name %s\n", n);
      return FAILED;
    }
  } else {
    i=currentdir;
    n="current dir";
  }
  d=opendir(i);
  if (!d) {
    fprintf(stderr,"ls: %s is not a directory\n", n);
    return FAILED;
  }
  /*  printf("Directory of %s, inode %u\n", n, i); */
  while (e=readdir(d)) {
    printf("%5u %20s ", e->inode, e->name);
    o=fetchinode(e->inode);
    switch(o->type) {
    case IN_TYPE_DEV:
      printf("-      dev %u/%u\n", o->first, o->len);
      break;
    case IN_TYPE_FILE:
      printf("%6u\n", o->len);
      break;
    case IN_TYPE_DIR:
      printf("%6u dir\n", o->len);
      break;
    default:
      printf("         unrecognised\n");
    }
  }
  closedir(d);
  return NULL;
}

/* --- builtins that might be seperate at some point --- */

int dl_main(argc,argv)
     int argc;
     char **argv;
{
  FILE *f;
  unsigned int len;

  if (!(f=fopen("/dev/serial","r"))) {
    return FAILED;
  } else {
    len=freadint(f);
    fprintf(stderr,"dl: getting %u bytes\n",len);
    while(len--) fputc(fgetc(f),stdout);
    fclose(f);
    return NULL;
  }
}

int cat_main(argc,argv)
     int argc;
     char **argv;
{
  FILE *fp;

  if (argc==1) filecopy(stdin);
  else
    while (--argc > 0)
      if ((fp=fopen(*++argv,"r"))==NULL) {
	fprintf(stderr,"cat: can't open %s\n", *argv);
	return FAILED;
      } else {
	filecopy(fp);
	fclose(fp);
      }
  return NULL;
}

void filecopy(fp)
     FILE *fp;
{
  int c;

  while ((c=getc(fp))!=EOF)
    putc(c,stdout);
}

/* --- sys functions --- */

int link(old, new) 
     char *old;
     char *new;
{
  int i;
  INODE *n;
  
  i=nametoinode(old);
  if (mkdirentry(new, i)) {
    return FAILED;
  } else {
    n=fetchinode(i);
    n->refcount++;
    return NULL;
  }
}

int mkfs() {
  int i, j, b;
  INODE *in;
  DIRENT *de;

  printf("Zen mkfs 0.2\n\n");
  printf("zeroing disk");
  for(j=0;j<NUMBLOCKS;j++) {
    putchar('.');
    getblock(j);
    for(i=0;i<BLOCKSIZE;i++) (currentbuffer->buf)[i]=0;
  }
  printf("\nmarking block table..\n");
  getblock(0);
  for(j=0;j<(InodeBlockNum(NUMINODES)+1);j++) (currentbuffer->buf)[j]=1;
  printf("creating root directory..\n");
  b=allocblock();
  printf("allocated block %u for root dir\n",b);
  printf("writing dummy inode\n");
  getblock(InodeBlockNum(0));
  in=(INODE *)((currentbuffer->buf)+InodeBlockIndex(0));
  in->type=IN_TYPE_FILE;
  printf("writing root inode\n");
  getblock(InodeBlockNum(1));
  in=(INODE *)((currentbuffer->buf)+InodeBlockIndex(1));
  in->type=IN_TYPE_DIR;
  in->refcount=1;
  in->first=b;
  in->len=1;
  printf("writing root selfref dirent\n");
  getblock(b);
  de=(DIRENT *)((currentbuffer->buf)+sizeof(BLOCK));
  de->inode=1;
  strcpy(de->name,"slash");
  printf("Making files...\n");
  mkdir("/dev");
  mkdev("/dev/console", DEV_CONSOLE, 0);
  mkdev("/dev/serial", DEV_SERIAL, 0);
  mkdev("/dev/null", DEV_CONST, '\0');
  mkdev("/dev/a", DEV_CONST, 'a');
  mkdir("/bin"); 
  mkdir("/tmp");
  ls("/");
  printf("mkfs done\n");
  return NULL;
}

int nametoinode(name)
     char *name;
{
  int d, l;
  DIR *r;
  DIRENT *e;
  char *k, c;

  if (!name) return 1;
  if (name[strlen(name)]=='/') name[strlen(name)]='\0';
  if (name[0]=='/') {
    d=1;
    name++;
  } else {
    d=currentdir;
  }
  while(*name) {
    if (*name=='/') name++;
    r=opendir(d);
    if (!r) return d; /* found in the form /path/to/file/rubbish;
			 we know it's the file as we can't indirect
			 through a file! */
    if (k=index(name, '/')) {
      l=k-name;
    } else {
      l=strlen(name);
    }
    c=name[l];
    name[l]='\0';
    while ((e=readdir(r)) && strcmp(name, e->name)) /* null body */;
    name[l]=c;
    closedir(r);
    if (!e) return NULL; /* file not found */
    d=e->inode;
    name += l;
  }
  return d;
}

char *dirname(name)
     char *name;
{
  int l;
  char *i;

  if (i=rindex(name, '/')) {
    l=i-name;
    if(l) {
      i=malloc(l+1);
      if (!i) return 0;
      strncpy(i,name,l);
      i[l]='\0';
    } else {
      i=malloc(2);
      if (!i) return 0;
      i[0]='/';
      i[1]='\0';
    }
  } else {
    i=malloc(1);
    if (!i) return 0;
    i[0]='\0';
  }
  return i;
}

char *basename(name)
     char *name;
{
  char *i, *j;

  if (!(i=rindex(name, '/'))) {
    i=name;
  } else {
    i++;
  }
  j=malloc(strlen(i)+1);
  if (!j) return 0;
  strcpy(j,i);
  return j;
}

/* This is a syscall, so I don't want to do it as a macro. (The
   implementation might change at a later date anyway.) */
int mkdir(name)
     char *name;
{
  return create(name, IN_TYPE_DIR);
}

/* Make a device file */
int mkdev(name, dev, unit)
     char *name;
     int dev;
     int unit;
{
  int inode;
  INODE *n;

  inode=allocinode(IN_TYPE_DEV);
  n=fetchinode(inode);
  n->refcount=1;
  n->first=dev;
  n->len=unit;
  if (mkdirentry(name, inode)) {
    freeinode(inode);
    return FAILED;
  } else {
    return NULL;
  }
}

/* Create a new entry with name *name, type t. */
int create(name, t)
     char *name;
     char t;
{
  int inode, block;
  INODE *n;

  block=allocblock();
  inode=allocinode(t);
  n=fetchinode(inode);
  n->refcount=1;
  n->first=block;
  n->len=0;
  if (mkdirentry(name, inode)) {
    freeinode(inode);
    freeblock(block);
    return FAILED;
  } else {
    return NULL;
  }
}

/* Make a directory entry with name name for inode #inode. */
int mkdirentry(name, inode)
     char *name;
     int inode;
{
  int b, l, s, nb, d;
  INODE *n;
  DIRENT *e;
  char *x;

  d=nametoinode(name);
  if (d) {
    fprintf(stderr,"mkdirentry: %s already exists (inode %u)\n",name,d);
    return FAILED;
  }
  x=dirname(name);
  d=nametoinode(x);
  free(x);
  if (!d) {
    fprintf(stderr,"mkdirentry: invalid path, no parent for %s\n",name);
    return FAILED;
  }
  x=basename(name);
  n=fetchinode(d);
  b=n->first;
  l=sizeof(BLOCK) + ((n->len++)*sizeof(DIRENT));
  getblock(b);
  while(l>BLOCKSIZE) {
    l -= BLOCKSIZE;
    b=((BLOCK *)currentbuffer->buf)->next;
    getblock(b);
  };
  e=(DIRENT *)malloc(sizeof(DIRENT));
  if (!e) return FAILED;
  e->inode=inode;
  strcpy(e->name,x);
  for(s=0;s<sizeof(DIRENT);s++) {
    (currentbuffer->buf)[l++]=*((char *)e+s);
    if (l>=BLOCKSIZE) {
      nb=allocblock();
      ((BLOCK *)currentbuffer->buf)->next=nb;
      getblock(nb);
      ((BLOCK *)currentbuffer->buf)->prev=b;
      l=sizeof(BLOCK);
    }
  }
  free((char *)e);
  free(x);
  return NULL;
}

void memcpy(dest, src, size)
     char *dest;
     char *src;
     int size;
{
  while(size--) *dest++=*src++;
}

int unlink(name) 
     char *name; 
{
  /* Strategy: read the last entry into the buffer and replace the one
     we want to delete with it, then shorten the directory, freeing
     the last block if necessary, then lessen the inode's refcount,
     and delete the file's data if required. 
  */
  char *p;
  int inode, flag, i, j, filenode;
  DIRENT *final, *e, *ee, *eee;
  DIR *dir, *mod;
  INODE *n;

  filenode=nametoinode(name);
  if (!filenode) {
    fprintf(stderr,"unlink: %s nonexistant\n",name);
    return FAILED;
  }
  n=fetchinode(filenode);
  if ((n->type)==IN_TYPE_DIR && (n->len) && (n->refcount<=1)) {
    fprintf(stderr,"unlink: %s is a directory with files in\n", name);
    return FAILED;
  }
  p=dirname(name);
  if (!p) return FAILED;
  inode=nametoinode(p);
  free(p);
  p=basename(name);
  if (!p) return FAILED;
  /* We should be worrying if inode==0 now */
  final=(DIRENT *)malloc(sizeof(DIRENT));
  if (!final) return FAILED;
  mod=opendir(inode); /* to get it allocated right */
  eee=mod->dirent; /* so we free it right later */
  e=ee=NULL;
  flag=1;
  dir=opendir(inode);
  do {
    ee=e;
    if (flag) memcpy((char *)mod, (char *)dir, sizeof(DIR));
    e=readdir(dir);
    if (e && (!strcmp(e->name,p))) flag=0;
  } while (e);
  mod->dirent=eee;
  free(p);
  /* don't need to check ee!=NULL, as we know the file exists */
  memcpy((char *)final, (char *)ee, sizeof(DIRENT));
  if ((dir->bpos)<(sizeof(BLOCK)+sizeof(DIRENT))) {
    /* I'm not convinced that the above test is correct. */
    getblock(dir->block);
    i=((BLOCK *)currentbuffer->buf)->prev;
    getblock(i);
    ((BLOCK *)currentbuffer->buf)->next=0;
    freeblock(dir->block);
  }
  closedir(dir);
  getblock(mod->block);
  p=(char *)final;
  for(i=0; i<sizeof(DIRENT); i++) {
    (currentbuffer->buf)[mod->bpos++]=p[i];
    if ((mod->bpos)>=BLOCKSIZE) {
      mod->block=((BLOCK *)currentbuffer->buf)->next;
      getblock(mod->block);
      mod->bpos=sizeof(BLOCK);
    }
  }
  free((char *)final);
  closedir(mod);
  n=fetchinode(inode);
  n->len--;
  /* Now fix the file's inode. */
  n=fetchinode(filenode);
  if (!(--n->refcount)) {
    /* Delete the file */
    i=n->first;
    do {
      getblock(i);
      j=((BLOCK *)currentbuffer->buf)->next;
      freeblock(i);
      i=j; /* 'cos freeblock smashes the buffer */
    } while (i);
    freeinode(filenode);
  };
  /* This is counterintuitive, but it will get written out on the next
     sync. Writeback caching is nice. */
  return NULL;
}

DIR *opendir(i)
     int i;
{
  DIR *d;
  INODE *n;

  n=fetchinode(i);
  if ((n->type)!=IN_TYPE_DIR) return NULL;
  d=(DIR *)malloc(sizeof(DIR));
  if (!d) return NULL;
  d->block=n->first;
  d->bpos=sizeof(BLOCK);
  d->fpos=0;
  d->len=n->len;
  d->dirent=(DIRENT *)malloc(sizeof(DIRENT));
  if (!d->dirent) {
    free((char *)d);
    return NULL;
  }
  return d;
}

void closedir(d)
     DIR *d;
{
  free((char *)d->dirent);
  free((char *)d);
}

DIRENT *readdir(d)
     DIR *d;
{
  char *p;
  int s;

  if (d->fpos >= d->len) return NULL;

  p=(char *)(d->dirent);
  getblock(d->block);
  for(s=0;s<sizeof(DIRENT);s++) {
    *(p++)=(currentbuffer->buf)[d->bpos++];
    if ((d->bpos)>=BLOCKSIZE) {
      d->block=((BLOCK *)currentbuffer->buf)->next;
      d->bpos=sizeof(BLOCK);
      getblock(d->block);
    }
  }
  d->fpos++;
  return d->dirent;
}

INODE *fetchinode(n)
     int n;
{
  getblock(InodeBlockNum(n));
  return (INODE *)((currentbuffer->buf)+InodeBlockIndex(n));
}

/* new inode of type t */
int allocinode(t)
     char t;
{
  int i;
  INODE *n;

  for(i=0;i<NUMINODES;i++) {
    n=fetchinode(i);
    if(!(n->type)) {
      n->type = t;
      return i;
    }
  }
  fprintf(stderr,"allocinode: disk full; no free inodes\n");
  return 0;
}

void freeinode(i)
     int i;
{
  INODE *n;

  n=fetchinode(i);
  n->type = IN_TYPE_NONE;
}

/* This is HORRIBLY inefficient. */
int allocblock() {
  int i;
  
  for(i=0;i<NUMBLOCKS;i++) {
    getblock(FreeBlockNum(i));
    if(!(currentbuffer->buf)[FreeBlockIndex(i)]) {
      (currentbuffer->buf)[FreeBlockIndex(i)]=1;
      return i;
    }
  }
  fprintf(stderr,"allocblock: disk full; no free blocks\n");
  return 0;
}

void freeblock(b)
     int b;
{
  getblock(FreeBlockNum(b));
  (currentbuffer->buf)[FreeBlockIndex(b)]=0;
}

/* -- */

int feof(f)
     FILE *f;
{
  return ((f->fpos)>=(f->len));
}

void fflush(f)
     FILE *f;
{
  /* do nothing -- files are not buffered */
}

FILE *fopen(name, mode) 
     char *name;
     char *mode;
{
  FILE *newfd;
  INODE *inode;
  int i, created;

  newfd = (FILE *)malloc(sizeof(FILE));
  if (!newfd) return NULL;
  created=0;
  if (!(i=nametoinode(name))) {
    /* file doesn't exist... yet */
    if (mode[0]=='r') {
      /* fprintf(stderr,"fopen: can't open %s for reading.\n", name); */
      free((char *)newfd);
      return NULL;
    }
    create(name, IN_TYPE_FILE);
    created=1;
  }
  if (mode[0]=='w') {
    i=nametoinode(name);
    if (i && (inode=fetchinode(i),inode->type==IN_TYPE_DEV)) {
      /* do nothing --- don't smash the device! */
    } else if (!created) {
      unlink(name);
      create(name, IN_TYPE_FILE);
    }
  }
  i=nametoinode(name);
  newfd->flags = 0;
  if (name) {
    inode=fetchinode(i);
    /* Note that you can open a directory; you just can't do anything
       with it except close it again. :-) */
    newfd->inode = i;
    newfd->type = inode->type; /* these work for devs and files */
    newfd->len = inode->len;
    newfd->block = inode->first;
    if (mode[0]=='a') {
      newfd->bpos = sizeof(BLOCK) + ((inode->len)%(BLOCKSIZE-sizeof(BLOCK)));
      newfd->fpos = inode->len;
    } else {
      newfd->bpos = sizeof(BLOCK);
      newfd->fpos = 0;
    }
  } else {
    /* fopen(NULL, NULL) opens the console initially */
    newfd->type = IN_TYPE_DEV;
    newfd->block = DEV_CONSOLE;
    newfd->len = 0;
    newfd->inode = NULL;
  }
  return newfd;
}

void fclose(fd)
     FILE *fd;
{
  INODE *i;

  if (!fd) return;
  if (fd->type == IN_TYPE_FILE) {
    i=fetchinode(fd->inode);
    i->len=fd->len;
  }
  free((char *)fd);
}

/* returns number of characters printed */
int fputn(n, f)
     long n;
     FILE *f;
{
  int i=0;

  if (n<0) {
    fputc('-', f);
    n = -n;
  }
  if (n>9) i=fputn(n/10, f);
  fputc('0'+(n%10), f);
  return i+1;
}

int fputu(n, f)
     unsigned int n;
     FILE *f;
{
  int i=0;

  if (n>9) i=fputu(n/10, f);
  fputc('0'+(n%10), f);
  return i+1;
}

int fputs(s, f)
     char *s;
     FILE *f;
{
  int count=0;

  while(*s) {
    count++;
    fputc(*s++, f);
  }
  return count;
}

void printf(s, q)
     char *s;
     int q;
{
  doprintf(stdout, s, &q);
}

void fprintf(f, s, q)
     FILE *f;
     char *s;
     int q;
{
  doprintf(f, s, &q);
}

/* Oversimplistic fprintf... but enough to be useful */
void doprintf(f, s, a) /* the prototype is ... */
     FILE *f;
     char *s;
     int *a; /* no, it's not! */
{
  int fw;

  while(*s) {
    if (*s=='%') {
      s++;
      fw=0;
      while (isnumber(*s)) fw = (10*fw) + (*s++) - '0';
      if (*s=='l') s++;
      switch(*s) {
      case 's':
	fw -= fputs((char *)*(a++), f);
	break;
      case 'd':
	fw -= fputn(*(a++), f);
	break;
      case 'u':
	fw -= fputu(*(a++), f);
	break;
      case 'c':
	fputc((char)*(a++), f);
	fw--;
	break;
      default: /* this copes with %% */
	fputc(*s, f);
	fw--;
	break;
      };
      if (fw>0) while (fw--) fputc(' ',f);
    } else fputc(*s, f);
    s++;
  };
}

char *fgets(s, n, f)
     char *s;
     int n;
     FILE *f;
{
  char *s1=s;
  char c;
  while(n-- && ( (c=fgetc(f)) != EOF ) && (*s++ = c)!='\n') {
    if ((c==8 || c==127) && s1!=s) {
      fputc(8, f);
      s--;
      n++;
    }
  }
  *s='\0';
  if ((!c)||(c==EOF)) return NULL;
  return(s1);
}

char fputc(c, f)
     FILE *f;
     char c;
{
  int ob;

  switch (f->type) {
  case IN_TYPE_DEV:
    switch (f->block) {
    case DEV_CONSOLE:
      if (c=='\r') c='\n';
      out(0, c);
#if PAGESIZE
      if ((c=='\n') || (pagercols > 79)) {
	pagercount--;
	pagercols = 0;
      } else pagercols++;
      if (!pagercount) {
	out(0, '<');
	out(0, 'm');
	out(0, '>');
	in(0);
	pagercount = PAGESIZE;
      }
#endif
      break;
    case DEV_CONST:
      /* bit bucket */
      break;
    };
    break;
  case IN_TYPE_FILE:
    getblock(f->block);
    (currentbuffer->buf)[f->bpos]=c;
    f->bpos++;
    f->fpos++;
    if (f->fpos >= f->len) f->len = f->fpos;
    if (f->bpos >= BLOCKSIZE) {
      ob=f->block;
      f->block = allocblock();
      getblock(ob);
      ((BLOCK *)(currentbuffer->buf))->next=f->block;
      getblock(f->block);
      ((BLOCK *)(currentbuffer->buf))->prev=ob;
      f->bpos = sizeof(BLOCK);
    }    
    break;
  };
  return c;
}

char fgetc(f)
     FILE *f;
{
  char c;

  switch (f->type) {
  case IN_TYPE_DEV:
    switch(f->block) {
    case DEV_CONSOLE:
      c=fputc(in(0),f);
      if (c==4) c='\0';
      return(c=='\r'?'\n':c);
    case DEV_SERIAL:
      return in(2);
    case DEV_CONST:
      return (char)(f->len);
    default:
      fprintf(stderr,"unknown device read\n");
      return EOF;
    }
  case IN_TYPE_FILE:
    if (f->fpos >= f->len) return EOF;
    getblock(f->block);
    c=(currentbuffer->buf)[f->bpos];
    f->bpos++;
    f->fpos++;
    if (f->bpos >= BLOCKSIZE) {
      f->block = ((BLOCK *)(currentbuffer->buf))->next;
      f->bpos = sizeof(BLOCK);
    }
    return c;
  }
  fprintf(stderr,"fgetc from non-entity %u\n",f);
  return NULL;
}

/* The K&R malloc routines. These are stolen straight from examples in
   K&R, and are extremely clever, which is why they are used here. As
   we don't have sbrk, I've modified them to create a real list at the
   start, and simply to fail when the list is exhausted.

   I have no idea how well these routines work. */

HEADER *base, *allocp = NULL;

/* My own routine, amount of RAM free in the pool */
int memfree() {
  HEADER *q;
  int f;

  f=0;
  q=allocp;
  do {
    f += q->s.size;
    q = q->s.ptr;
  } while (q!=allocp);
  return f*sizeof(HEADER);
}

void memclr(m,l)
     register char *m;
     register int l;
{
  while (l--) *m++=0;
}

char *malloc(nbytes)
     unsigned nbytes;
{
  HEADER *p, *q;
  int nunits;

#if ALLOCTRACK
  fprintf(stderr,"[a%u@",nbytes);
#endif
  nunits = 1+(nbytes+sizeof(HEADER)-1)/sizeof(HEADER);
  if ((q=allocp)==NULL) {
    base = (HEADER *)ramstart;
    allocp = q = base->s.ptr = base;
    base->s.size = (MEMORYTOP-ramstart)/sizeof(HEADER);
  }
  for(p=q->s.ptr;;q=p,p=p->s.ptr) {
    if (p->s.size >= nunits) {
      if (p->s.size == nunits)
	q->s.ptr=p->s.ptr;
      else {
	p->s.size -= nunits;
	p += p->s.size;
	p->s.size=nunits;
      }
      allocp=q;
      /* I give up. Let's BE a VAX. */
      memclr((char *)(p+1),nbytes);
#if ALLOCTRACK
      fprintf(stderr,"%u] ",(p+1));
#endif
      return((char *)(p+1));
    }
    if (p==allocp) return NULL; /* we don't have morecore */
  }
}

/* This *must* win some kind of award for "most confusing use of C
   pointers". *And* the original didn't check for free(NULL)! */
void free(ap)
     char *ap;
{
  HEADER *p, *q;
  
#if ALLOCTRACK
  fprintf(stderr,"[f%u] ",ap);
#endif
  if (!ap) return;
  p=(HEADER *)ap-1;
  for(q=allocp; !(p>q && p< q->s.ptr); q=q->s.ptr)
    if (q >= q->s.ptr && (p>q||p<q->s.ptr)) break;
  if (p+p->s.size==q->s.ptr) {
    p->s.size += q->s.ptr->s.size;
    p->s.ptr = q->s.ptr->s.ptr;
  } else p->s.ptr = q->s.ptr;
  if (q+q->s.size == p) {
    q->s.size += p->s.size;
    q->s.ptr = p->s.ptr;
  } else q->s.ptr = p;
  allocp = q;
}


/* Disk drivers */

void getblock(b)
     int b;
{
  BUFFER *u, *oldest;
  int age;

  if ((currentbuffer->block)==b) return;

  age = 0;
  u = oldest = currentbuffer;
  do {
    if (u->age > age) {
      oldest = u;
      age = u->age;
    }
    if (u->block == b) {
#if BUFTRACK
      fprintf(stderr,"[b%u=] ", (int)u);
#endif
      currentbuffer = u;
      return;
    }
  } while ((u=u->next) != currentbuffer);

#if BUFTRACK
  fprintf(stderr,"[b%u<%u] ", (int)oldest, b);
#endif
  writeblock(oldest->block, oldest->buf); /* sync */
  readblock(b, oldest->buf);
  u = currentbuffer;
  do {u->age++;} while((u=u->next) != currentbuffer);
  oldest->age = 0;
  oldest->block = b;
  currentbuffer = oldest;
}

void sync()
{
  BUFFER *u;

  u = currentbuffer;
  do {
#if BUFTRACK
    fprintf(stderr,"[b%u>%u] ",(int)u,u->block);
#endif
    writeblock(u->block, u->buf);
  } while ((u = u->next) != currentbuffer);
}

/* Low-level disk drivers */

void readblock(b, buf)
     int b;
     char *buf;
{
#if DISKTRACK
  fprintf(stderr,"[d<%u] ", b);
#endif
  out(0x10, b&255);
  out(0x11, b/256);
  out(0x12, (int)buf&255);
  out(0x13, ((int)buf)>>8);
  in(0x14);
}

void writeblock(b, buf)
     int b;
     char *buf;
{
#if DISKTRACK
  fprintf(stderr,"[d>%u] ", b);
#endif
  out(0x10, b&255);
  out(0x11, b/256);
  out(0x12, (int)buf&255);
  out(0x13, ((int)buf)>>8);
  out(0x14, 1);
}

